import { createElement, useEffect, useState } from 'react';
import { isObservable } from 'rxjs';
import { distinctUntilChanged, takeUntil } from 'rxjs/operators';
import { $ } from './fragment';
import { useDestroyObservable, useEmptyObject } from './shared';
export function createElement$(element) {
    return function element$(props) {
        // state for renderable props
        var _a = useState(Object.create(null)), streamProps = _a[0], setStreamProps = _a[1];
        // placeholder for Observable.error case -- will use it to rethrow error
        var _b = useState(null), error = _b[0], setError = _b[1];
        var destroy$ = useDestroyObservable();
        // store prev props to compare
        var _prevStreamProps = useEmptyObject();
        // keep subscriptions to unsubscribe on dynamic prop update
        var _subs = useEmptyObject();
        useEffect(function () {
            var isDirty = false;
            // check for obsolete props
            var delProps = Object.create(null);
            Object.keys(_prevStreamProps).forEach(function (key) {
                if (key in props) {
                    return;
                }
                isDirty = true;
                delete _prevStreamProps[key];
                // if previous property was Observable
                // kill subscription
                cleanSubscription(_subs, key);
                // remove from static props
                delProps[key] = void 0;
            });
            // update/track new props
            var nextProps = Object.create(null);
            var streamKeys = [];
            Object.keys(props).forEach(function (key) {
                // children are covered via <$>
                if (key == 'children') {
                    return;
                }
                var value = props[key];
                var prevValue = _prevStreamProps[key];
                var equal = Object.is(value, prevValue);
                if (equal) {
                    return;
                }
                // if property changes and previous was Observable
                // we need to kill subscription
                cleanSubscription(_subs, key);
                // observable input params are added to observation
                // all static props are directly updated
                if (isObservable(value)) {
                    isDirty = true;
                    _prevStreamProps[key] = value;
                    nextProps[key] = void 0; // reset prev prop when new one is Observable
                    streamKeys.push(key);
                }
                else {
                    // forget outdated prev props
                    delete _prevStreamProps[key];
                }
            });
            // subscribe to new streams
            // some values might be received in sync way: like `of(…)` or `startWith(…)`
            // to optimize this we update all syncronously received values in one
            // commit to the state
            // {{{
            var isSync = true;
            streamKeys.forEach(function (key) {
                _subs[key] = props[key]
                    .pipe(distinctUntilChanged(), takeUntil(destroy$))
                    .subscribe({
                    // on value updates -- update props
                    next: function (data) {
                        // set sync values
                        if (isSync) {
                            isDirty = true;
                            nextProps[key] = data;
                        }
                        else {
                            // async set values
                            setStreamProps(function (p) {
                                var _a;
                                return Object.assign({}, p, (_a = {}, _a[key] = data, _a));
                            });
                        }
                    },
                    // on error -- rethrow error
                    error: function (error) {
                        setError({ error: error });
                    }
                    // on complete we just keep using accumulated value
                });
            });
            isSync = false;
            // }}}
            // remove obsolete props
            // & update static props
            if (isDirty) {
                setStreamProps(function (p) { return Object.assign({}, p, delProps, nextProps); });
            }
        }, [props]);
        // if error -- throw
        if (error) {
            throw error.error;
        }
        // using statically available props in render phase
        // so that `<$a alt="hello" href={ stream$ } >…</a>`
        // renders `<a alt="hello">…</a>` immediately
        var derivedProps = Object.keys(props).reduce(function (p, key) {
            var _a;
            if (isObservable(props[key])) {
                // ensure controlled elements stay controlled
                // if value is present and is not nullish
                // we make the input controlled
                if (key == 'value'
                    && (element == 'input'
                        && props.type != 'file'
                        && streamProps.type != 'file'
                        || element == 'select'
                        || element == 'textarea')) {
                    p[key] = (_a = streamProps.value) !== null && _a !== void 0 ? _a : '';
                }
                else {
                    p[key] = streamProps[key];
                }
            }
            else {
                p[key] = props[key];
            }
            return p;
        }, Object.create(null));
        return createElement(element, derivedProps, 
        // if children exist
        // they might be observable
        // so we pass em to <$> fragment
        // NOTE: children might not exist for elements like <input />
        props.children
            ? createElement($, null, props.children)
            : null);
    };
}
// helpers
function cleanSubscription(store, key) {
    if (store[key]) {
        store[key].unsubscribe();
        delete store[key];
    }
}
